import json
import logging
import signal
import sys

import daemon
import lockfile
from daemon import pidfile
from django.conf import settings
from django.core.management.base import BaseCommand
from pika_context_manager import PCM

from greg.models import Notification
from greg.utils import camel_to_snake

logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger()


def exception_handler(ex_cls, ex, tb):
    logger.critical("Uncaught exception:", exc_info=(ex_cls, ex, tb))


sys.excepthook = exception_handler


def get_notifications(limit):
    return Notification.objects.all()[:limit]


def generate_event_type(n: Notification) -> str:
    """
    Generate CloudEvent compliant type field.

    This can be used as routing key as well so that a future change to amqp where
    routing key is removed results in minor changes to the code.
    """
    object_type = camel_to_snake(n.object_type)
    return f"{settings.INTERNAL_RK_PREFIX}.{object_type}.{n.operation}"


def create_cloud_event_payload(n: Notification) -> str:
    """
    Produce a notification payload from a Notification object that matches CloudEvents
    spec.

    This ensures that the combination of source and id is unique for all messages
    published from this instance of greg. If you mix messages from multiple instances of
    greg this will fail.

    Using the object_type and the identifier gives us which api endpoint to use and
    which id to look up.

    specversion specifies version of CloudEvents spec, in our case 1.0. Note we're using
    only major and minor versions. This function was written at the time of specversion
    1.0.1.

    The type describes which sort of event has happened to the object identifier by id
    and source, i.e no.local.greg.person.add
    """

    object_type = camel_to_snake(n.object_type)
    return json.dumps(
        {
            "id": str(n.id),
            "source": f"urn:greg:{settings.INSTANCE_NAME}:{object_type}:{n.identifier}",
            "specversion": "1.0",
            "type": generate_event_type(n),
        }
    )


def handle_one_notification(notification: Notification, pcm: PCM, exchange: str):
    """
    Publish a message to an exchange with the information contained in a Notification
    and delete the Notification afterwards.
    """
    payload = create_cloud_event_payload(notification)
    # Use event_type for routing_key so that a future switch to amqp 1.0 that does not
    # have routing keys is easier
    routing_key = generate_event_type(notification)

    if pcm.publish(
        exchange,
        routing_key,
        payload,
        content_type="application/json",
    ):
        logger.info(
            "Published message %s with routing key %s to exchange %s",
            payload,
            routing_key,
            exchange,
        )

        notification.delete()
    else:
        logger.error(
            "Failed published message %s with routing key %s to exchange %s",
            payload,
            routing_key,
            exchange,
        )


class RunState:
    def __init__(self):
        self.run = True

    def stop(self, *args):
        self.run = False


class Command(BaseCommand):
    help = "Publish notifications via AMQP"

    def add_arguments(self, parser):
        parser.add_argument(
            "--detach", action="store_true", help="Run detached as a dæmon"
        )
        parser.add_argument(
            "--use-pidfile", action="store_true", help="Use a PID lockfile"
        )

    def handle(self, *args, **options):
        state = RunState()
        lock_context = None
        if options.get("use_pidfile"):
            lock_context = pidfile.TimeoutPIDLockFile(
                settings.NOTIFICATION_PUBLISHER["daemon"]["pid_file"]
            )

        try:
            with daemon.DaemonContext(
                pidfile=lock_context,
                files_preserve=[x.stream.fileno() for x in logger.handlers],
                stderr=sys.stderr,
                stdout=sys.stdout,
                signal_map={signal.SIGINT: state.stop, signal.SIGTERM: state.stop},
                detach_process=options.get("detach"),
            ):
                logger.info(
                    "Running %s", "detached" if options.get("detach") else "attached"
                )
                with PCM(**settings.NOTIFICATION_PUBLISHER["mq"]["connection"]) as pcm:
                    exchange = settings.NOTIFICATION_PUBLISHER["mq"]["exchange"]
                    while state.run:
                        for notification in get_notifications(limit=500):
                            handle_one_notification(
                                notification=notification, pcm=pcm, exchange=exchange
                            )

                        pcm.sleep(
                            settings.NOTIFICATION_PUBLISHER["daemon"]["poll_interval"]
                        )
        except lockfile.AlreadyLocked as e:
            logger.warning("Can't start daemon: %s", e)
            # TODO: Figure out how to emit a code other than 0 here
            # TODO: Also, echo the error to stderr
        finally:
            logger.info("Stopped")
