import logging.config
import signal
import sys
from threading import Event

import daemon
import lockfile
from daemon import pidfile
from django.conf import settings
from django.core.management.base import BaseCommand

from greg.schedule import ExpiringRolesNotification

logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger()


def exception_handler(ex_cls, ex, tb):
    logger.critical("Uncaught exception:", exc_info=(ex_cls, ex, tb))


sys.excepthook = exception_handler


class Command(BaseCommand):
    """
    This command starts a basic task runner. All tasks it is supposed to
    run are given explicitly in code
    """

    help = "Start schedule task runner"

    def add_arguments(self, parser):
        parser.add_argument(
            "--detach", action="store_true", help="Run detached as a dæmon"
        )
        parser.add_argument(
            "--use-pidfile", action="store_true", help="Use a PID lockfile"
        )

    def handle(self, *args, **options):
        lock_context = None
        if options.get("use_pidfile"):
            lock_context = pidfile.TimeoutPIDLockFile(
                settings.SCHEDULE_TASKS["daemon"]["pid_file"]
            )

        try:
            exit_event = Event()

            def exit_wrapper(*args):
                exit_event.set()

            with daemon.DaemonContext(
                pidfile=lock_context,
                files_preserve=[x.stream.fileno() for x in logger.handlers],
                stderr=sys.stderr,
                stdout=sys.stdout,
                signal_map={signal.SIGINT: exit_wrapper, signal.SIGTERM: exit_wrapper},
                detach_process=options.get("detach"),
            ):
                logger.info(
                    "Running %s", "detached" if options.get("detach") else "attached"
                )

                # For now there is just one task, but the idea is that more can
                # be added
                expiring_role_notification_task = ExpiringRolesNotification()

                while not exit_event.is_set():
                    expiring_role_notification_task.run()
                    # The single task is set up for far only needs to run once
                    # a day, but running every 6 hours just in case a run fails,
                    # it is up the task it self to figure out if needs to do
                    # something every time it is called
                    exit_event.wait(60 * 60 * 6)

        except lockfile.AlreadyLocked as e:
            logger.warning("Can't start daemon: %s", e)
            sys.exit(1)
        finally:
            logger.info("Stopped")
